package tests

import (
	"fmt"
	"os"
	"path"
	"testing"

	"github.com/davecgh/go-spew/spew"
	_ "github.com/nspcc-dev/neo-go/pkg/compiler"
	"github.com/nspcc-dev/neo-go/pkg/neotest"
	"github.com/nspcc-dev/neo-go/pkg/neotest/chain"
	"github.com/nspcc-dev/neo-go/pkg/smartcontract/callflag"
	"github.com/nspcc-dev/neo-go/pkg/vm"
	"github.com/nspcc-dev/neo-go/pkg/vm/stackitem"
)

const ctrPath = "../contract"

// Key for tests
var (
	validKey   = []byte{1, 2, 3, 4, 5}
	invalidKey = []byte{1, 2, 3}
)

func newExecutor(t *testing.T) *neotest.Executor {
	bc, acc := chain.NewSingle(t)
	return neotest.NewExecutor(t, bc, acc, acc)
}

func TestContract_GetNumber(t *testing.T) {
	e := newExecutor(t)
	ctrGetNum := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrGetNum, nil)
	inv := e.CommitteeInvoker(ctrGetNum.Hash)
	inv.InvokeFail(t, "Cannot get number", "getNumber", validKey)
}

func TestContract_PutNumber(t *testing.T) {
	e := newExecutor(t)
	ctrPutGetNum := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrPutGetNum, nil)
	inv := e.CommitteeInvoker(ctrPutGetNum.Hash)
	inv.Invoke(t, stackitem.Null{}, "putNumber", validKey, 42)

	inv.Invoke(t, 42, "getNumber", validKey)
}

func TestContract_InvalidKey(t *testing.T) {
	e := newExecutor(t)
	ctrPutGetNum := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrPutGetNum, nil)
	inv := e.CommitteeInvoker(ctrPutGetNum.Hash)
	inv.InvokeFail(t, "Invalid key size", "putNumber", invalidKey, 42)
	inv.InvokeFail(t, "Invalid key size", "getNumber", invalidKey)
}

func TestContract_Add(t *testing.T) {
	e := newExecutor(t)
	ctrAdd := neotest.CompileFile(t, e.CommitteeHash, ctrPath, path.Join(ctrPath, "neo-go.yml"))
	e.DeployContract(t, ctrAdd, nil)
	spew.Dump(ctrAdd.Manifest)

	govm := vm.New()
	govm.LoadNEFMethod(ctrAdd.NEF, ctrAdd.Hash, ctrAdd.Hash, callflag.All, true, 170, -1, nil)
	govm.Context().Estack().PushVal(1)
	govm.Context().Estack().PushVal(3)
	//fmt.Println(govm.Context().CurrInstr())
	_, curInstr := govm.Context().CurrInstr()
	//while curInstr != RET do
	for curInstr != 0x40 {
		fmt.Println(curInstr)
		govm.Step()
		_, curInstr = govm.Context().CurrInstr()
	}
	fmt.Println(govm.Context().CurrInstr())

	//govm.DumpEStack()
	//govm.Run()
	res := govm.Context().Estack().Pop().Value()
	fmt.Println(res)

	// manif, _ := json.MarshalIndent(ctrAdd.Manifest, "", " ")
	// fmt.Println(string(manif))
	inv := e.CommitteeInvoker(ctrAdd.Hash)
	inv.Invoke(t, 3, "add", 1, 2)
}

func Test(t *testing.T) {
	govm := vm.New()
	err := govm.LoadFileWithFlags(path.Join(ctrPath, "contract.nef"), callflag.All)
	if err != nil {
		fmt.Fprintln(os.Stderr, "NEF loading failed")
		t.FailNow()
	}
	govm.PrintOps(os.Stdout)
	//step in a loop for debug
	//
	govm.Context().Jump(170)
	govm.Context().Estack().PushVal(1)
	govm.Context().Estack().PushVal(2)

	govm.Run()
	res := govm.Context().Estack().Pop().Value()
	fmt.Println(res)
}
